// -*- C++ -*-
/*!
 * @file  Navigation.cpp
 * @brief compornent
 * $Date$
 *
 * $Id$
 */

#include "Navigation.h"
#include "PathType.h"
#include<math.h>

// Module specification
// <rtc-template block="module_spec">
static const char* navigation_spec[] =
  {
    "implementation_id", "Navigation",
    "type_name",         "Navigation",
    "description",       "compornent",
    "version",           "1.0",
    "vendor",            "MyName",
    "category",          "example",
    "activity_type",     "DataFlowComponent",
    "max_instance",      "10",
    "language",          "C++",
    "lang_type",         "compile",
    // Configuration variables
    "conf.default.judge_radius", "0.5",
    "conf.default.max_vel", "0.3",

    ""
  };
// </rtc-template>


enum{
  FINISH = 0,
  ANGLE_CHECK,
  TURNING,
  TURNING2,
  GOAL_CHECK,
  SPEED_DOWN,
  STOPPING
};
Navigation::Navigation(RTC::Manager* manager)
  : RTC::DataFlowComponentBase(manager),
    // <rtc-template block="initializer">
    m_positionIn("position", m_position),
    m_min_pathIn("min_path", m_min_path),
    m_pathOut("path", m_path),
    m_max_speedOut("max_speed", m_max_speed),
    m_max_accelerationOut("max_acceleration", m_max_acceleration),
    
    // </rtc-template>
	dummy(0)
{
  // Registration: InPort/OutPort/Service
  // <rtc-template block="registration">
  // Set InPort buffers
  registerInPort("position", m_positionIn);
  registerInPort("min_path", m_min_pathIn);
  
  // Set OutPort buffer
  registerOutPort("path", m_pathOut);
  registerOutPort("max_speed", m_max_speedOut);
  registerOutPort("max_acceleration", m_max_accelerationOut);
  
  // Set service provider to Ports
  
  // Set service consumers to Ports
  
  // Set CORBA Service Ports
  
  // </rtc-template>

}

Navigation::~Navigation()
{
}


RTC::ReturnCode_t Navigation::onInitialize()
{
  // <rtc-template block="bind_config">
  // Bind variables and configuration variable
  bindParameter("judge_radius", m_judge_radius, "0.5");
  bindParameter("max_vel", m_max_vel, "0.3");

  printf("Navigation Component ver.0.1\n Initialised\n\n");  
  // </rtc-template>
  return RTC::RTC_OK;
}



/*
RTC::ReturnCode_t Navigation::onFinalize()
{
  return RTC::RTC_OK;
}
*/

/*
RTC::ReturnCode_t Navigation::onStartup(RTC::UniqueId ec_id)
{
  return RTC::RTC_OK;
}
*/

/*
RTC::ReturnCode_t Navigation::onShutdown(RTC::UniqueId ec_id)
{
  return RTC::RTC_OK;
}
*/


RTC::ReturnCode_t Navigation::onActivated(RTC::UniqueId ec_id)
{
    return RTC::RTC_OK;
}


/*
RTC::ReturnCode_t Navigation::onDeactivated(RTC::UniqueId ec_id)
{
  return RTC::RTC_OK;
}
*/

double angle_diff(double theta, double target){
  double theta_diff;
  theta_diff = theta-target;

  while(theta_diff < -M_PI)theta_diff+=2*M_PI;
  while(theta_diff > M_PI)theta_diff-=2*M_PI;

  return fabs(theta_diff);
}

double distance(double x, double y, double tx,double ty){
  return sqrt((x-tx)*(x-tx)+(y-ty)*(y-ty));
}

double over_line(double x,double y, double tx, double ty,double ttheta){
  double rx,ry;

  rx = (x-tx)*cos(-ttheta)-(y-ty)*sin(-ttheta);
  ry = (x-tx)*sin(-ttheta)+(y-ty)*cos(-ttheta);

  return rx;
}

RTC::ReturnCode_t Navigation::onExecute(RTC::UniqueId ec_id)
{
int i;
 static double next_angle; 
  /**/

 if(!m_min_pathIn.isEmpty()){
   while(!m_min_pathIn.isEmpty())
     m_min_pathIn.read();
   
    path_num = m_min_path.pose.length();
    state = ANGLE_CHECK;
    /*if(path_NUM>=2)
      next_angle = 
	atan2(m_min_path.pose[1].position.y-m_min_path.pose[0].position.y,
	      m_min_path.pose[1].position.x-m_min_path.pose[0].position.x);
    else
      next_angle = 0;
    */
    current_id = 0;
    for(i = 0;i < path_num;i++){
      printf("%d %f %f %f\n",i, 
	     m_min_path.pose[i].position.x,
	     m_min_path.pose[i].position.y,
	     m_min_path.pose[i].heading);
    }
  }

  /*If odometry is updated*/
  if(!m_positionIn.isEmpty()){
    while(!m_positionIn.isEmpty())
      m_positionIn.read();

    switch(state){
    case ANGLE_CHECK:
      if(current_id != path_num-1){
	current_id++;
	next_angle = 
	  atan2(m_min_path.pose[current_id].position.y
		-m_min_path.pose[current_id-1].position.y,
		m_min_path.pose[current_id].position.x
		-m_min_path.pose[current_id-1].position.x);
	if(angle_diff(next_angle,//m_min_path.pose[current_id].heading,
		      m_position.data.heading)> M_PI/4
	   /*&& current_id != 0*/){
	
	  
	  m_path.type = RUN_SPIN;
	  m_path.theta = next_angle;//m_min_path.pose[current_id].heading;
	  m_path.w = 1;
	  m_pathOut.write();
	  state = TURNING;
	  
	  printf("turn\n");
	}else{
	  m_path.type = RUN_SPIN;
	  m_path.theta = next_angle;//m_min_path.pose[current_id].heading;
	  m_path.w = 1;
	  m_pathOut.write();
	  state = GOAL_CHECK;
	  printf("no turn \n");
	}
      }else{/*last*/
	m_path.type = RUN_SPIN;
	m_path.theta = m_min_path.pose[current_id].heading;
	m_path.w = 1;
	m_pathOut.write();
	state = TURNING2;
      }
      break;

    case TURNING:
      if(angle_diff(next_angle,//m_min_path.pose[current_id].heading,
		    m_position.data.heading)< M_PI/10){
	printf("turned\n");
	state = GOAL_CHECK;
      }
      break;

    case TURNING2:
      if(angle_diff(m_min_path.pose[current_id].heading,
		    m_position.data.heading)< 0.02){
	m_path.type = RUN_STOP;
	m_pathOut.write();
	state = FINISH;
	printf("goal\n");
      }
      break;

    case GOAL_CHECK:
      
      // current_id++;
	 m_path.type = RUN_LINEFOLLOW;
	 m_path.x = m_min_path.pose[current_id].position.x;
	 m_path.y = m_min_path.pose[current_id].position.y;
	
	 m_path.theta = next_angle;
	 m_path.v = m_max_vel;
	 m_pathOut.write();
	 state = SPEED_DOWN;
	 printf("line follow\n");
      
	break;

    case SPEED_DOWN:
      //       if(distance(m_min_path.pose_list[current_id].x, m_min_path.pose_list[current_id].y,
      //	 m_position.x, m_position.y)< 1.2){
      if(over_line(m_position.data.position.x, m_position.data.position.y,
		   m_min_path.pose[current_id].position.x, 
		   m_min_path.pose[current_id].position.y,
		   m_min_path.pose[current_id].heading)>-m_judge_radius){
	m_path.type = RUN_LINEFOLLOW;
	m_path.x = m_min_path.pose[current_id].position.x;
	m_path.y = m_min_path.pose[current_id].position.y;
	m_path.theta = next_angle;//m_min_path.pose[current_id].heading;
	m_path.v = 0.2;
	m_pathOut.write();
	state = STOPPING;
	printf("stopping\n");
      }
      break;

    case STOPPING:
	if(over_line(m_position.data.position.x,m_position.data.position.y,
		     m_min_path.pose[current_id].position.x, 
		     m_min_path.pose[current_id].position.y,
		     next_angle//m_min_path.pose[current_id].heading
		     )>0){
	  
	m_path.type = RUN_STOP;
	m_pathOut.write();
	state = ANGLE_CHECK;
	//current_id++;
	printf("stop\n");
      }else{
	printf("reach %f\n",distance(m_min_path.pose[current_id].position.x, 
				     m_min_path.pose[current_id].position.y,
		  m_position.data.position.x, m_position.data.position.y));    
      }
      break;

    case FINISH:

      break;
    }
    //    printf(".\n");
  }

  return RTC::RTC_OK;
}


/*
RTC::ReturnCode_t Navigation::onAborting(RTC::UniqueId ec_id)
{
  return RTC::RTC_OK;
}
*/

/*
RTC::ReturnCode_t Navigation::onError(RTC::UniqueId ec_id)
{
  return RTC::RTC_OK;
}
*/

/*
RTC::ReturnCode_t Navigation::onReset(RTC::UniqueId ec_id)
{
  return RTC::RTC_OK;
}
*/

/*
RTC::ReturnCode_t Navigation::onStateUpdate(RTC::UniqueId ec_id)
{
  return RTC::RTC_OK;
}
*/

/*
RTC::ReturnCode_t Navigation::onRateChanged(RTC::UniqueId ec_id)
{
  return RTC::RTC_OK;
}
*/



extern "C"
{

  void NavigationInit(RTC::Manager* manager)
  {
    RTC::Properties profile(navigation_spec);
    manager->registerFactory(profile,
                             RTC::Create<Navigation>,
                             RTC::Delete<Navigation>);
  }

};


